import os
import random
from pathlib import Path
from typing import Union, Any

from fameio.tools import keys_to_lower

from scengen.cli import CreateOptions
from scengen.generation.misc import get_all_ids_from, create_new_unique_id, cast_numeric_strings, \
    get_relative_paths_in_dir, extract_numbers_from_string
from scengen.logs import log, log_and_raise_critical

numeric = Union[int, float]


SEPARATOR = ";"
RANGE_INT_IDENTIFIER = "range_int"
RANGE_FLOAT_IDENTIFIER = "range_float"
RANGE_IDENTIFIER_DEPRECATED = "range("
CHOOSE_IDENTIFIER = "choose"
PICKFILE_IDENTIFIER = "pickfile"
REPLACEMENT_IDENTIFIER = "//"
KEY_THIS_AGENT = f"{REPLACEMENT_IDENTIFIER}THIS_AGENT"


ERR_INVALID_RANGE_INPUT = (
    "Received invalid range input in form '{}'. Please provide in format "
    f"'{RANGE_INT_IDENTIFIER}(minimum_integer, maximum_integer)' or "
    f"'{RANGE_FLOAT_IDENTIFIER}(minimum_float, maximum_float)'. "
    "Negative integers are considered {}"
)
ERR_INVALID_RANGE_ORDER = "Received invalid range input in form '{}'. First value must be larger equal to second value."
ERR_MISSING_MATCH = "Found identifier(s) '{}' in `external_ids` but could not match with given agent(s) '{}'."
ERR_STATIC_CONTRACT = ("Expected dynamic, but found static contract template instead. Replace either `senderid` or "
                       "`receiverid` with a dynamic replacement_identifier in contract '{}'")
WARN_ROUNDED_NUMBER_OF_AGENTS = ("Rounded `agent_count` to '{}'. Make sure you provide a single integer or "
                                 "'{}' instead of '{}'.")
ERR_DEPRECATED_RANGE_IDENTIFIER = "Found deprecated identifier in '{}'. Please use '{}' or '{}' instead."
ERR_NO_INTEGER = "Expected a single integer or '{}' but received '{}' for `agent_count` instead."
DEBUG_NO_PATH_TO_BE_REPLACED_IN = "No path to be replaced for Attribute '{}: {}'."
ERR_COULD_NOT_MAP_RANGE_VALUES = "Could not map range values '{}' to minimum, maximum values."
ERR_FAILED_RESOLVE_ID = ("Cannot match replacement Identifier '{}' from Contract '{}' to any existing Agent. "
                         f"Make sure to reference either '{KEY_THIS_AGENT}' "
                         f"or any dynamically created agent.")


def get_number_of_agents_to_create(agent_count: Union[list[Any], Any], options: dict) -> int:
    """
    Returns an integer number from field `agent_count`
    Accepts float values by rounding to integer, but raises warning to notify user about wrong type,
    for any other data type an Error is raised
    """
    value_from_field = get_value_from_field(agent_count, options, allow_negative=False)
    if not isinstance(value_from_field, int):
        try:
            rounded_number = round(value_from_field)
            log().warning(WARN_ROUNDED_NUMBER_OF_AGENTS.format(rounded_number, RANGE_INT_IDENTIFIER, agent_count))
            value_from_field = rounded_number
        except TypeError:
            log_and_raise_critical(ERR_NO_INTEGER.format(RANGE_INT_IDENTIFIER, agent_count))
    return value_from_field


def get_value_from_field(input_value: Union[list[Any], Any], options: dict, allow_negative: bool = True) -> Any:
    """
    Returns value stored in `input_value` based on the user specification 'RANGE_INT_IDENTIFIER',
    'RANGE_INT_IDENTIFIER', 'CHOOSE_IDENTIFIER', 'PICKFILE_IDENTIFIER' or else just `input_value`.
    In range options, `allow_negative` is True as default but can limit allowed range to values >=0.
    In option `PICKFILE_IDENTIFIER`, `options[CreateOptions.DIRECTORY]` is used to get files: paths relative to scenario
    """
    if isinstance(input_value, str):
        if RANGE_IDENTIFIER_DEPRECATED in input_value.lower():
            log_and_raise_critical(
                ERR_DEPRECATED_RANGE_IDENTIFIER.format(input_value, RANGE_INT_IDENTIFIER, RANGE_FLOAT_IDENTIFIER)
            )
        elif RANGE_INT_IDENTIFIER in input_value.lower():
            input_range = digest_int_range(input_value)
            validate_input_range(input_range, allow_negative)
            value = random.randint(*input_range)
            log().debug(f"Chose random value '{value}' from '{input_value}'.")
        elif RANGE_FLOAT_IDENTIFIER in input_value.lower():
            input_range = digest_float_range(input_value)
            validate_input_range(input_range, allow_negative)
            value = random.uniform(*input_range)
            log().debug(f"Chose random value '{value}' from '{input_value}'.")
        elif CHOOSE_IDENTIFIER in input_value.lower():
            to_choose = digest_choose(input_value)
            value = random.choice(to_choose)
            log().debug(f"Chose random value '{value}' from list '{input_value}'.")
        elif PICKFILE_IDENTIFIER in input_value.lower():
            to_pick = digest_pickfile(input_value, options[CreateOptions.DIRECTORY])
            value = random.choice(to_pick)
            log().debug(f"Chose random file '{value}' from path '{input_value}'.")
        else:
            value = input_value
            log().debug(f"Received exactly one input value '{input_value}'.")
    else:
        value = input_value
        log().debug(f"Received exactly one input value '{input_value}'.")
    return value


def get_agent_id(agent_name: str, agent_number: int, n_of_agents_to_create: int) -> str:
    """
    Returns `agent_id` with leading REPLACEMENT_IDENTIFIER for `agent_name` considering its `agent_number`
    and `n_of_agents_to_create`
    """
    agent_id = REPLACEMENT_IDENTIFIER + agent_name
    agent_id += str(agent_number) if n_of_agents_to_create > 1 else ""
    return agent_id


def update_series_paths(scenario: dict, options: dict, template_dir: Path) -> None:
    """
    Appends relative paths directing to `template_dir` for CSV files defined in `scenario`s `Agents`
    and (optional) `StringSets`
    """
    config_dir = Path(options[CreateOptions.CONFIG]).parent
    output_dir = Path(options[CreateOptions.DIRECTORY])
    path_to_append = Path(os.path.relpath(config_dir, start=output_dir), template_dir.parent)
    for agent in scenario["Agents"]:
        agent = keys_to_lower(agent)
        replace_timeseries_path_in(agent.get("Attributes".lower(), {}), path_to_append)


def replace_timeseries_path_in(attributes: dict, template_path: Path) -> None:
    """Recursively modify timeseries path in-place in given `attributes` to link to `template_path`"""
    for attribute_name, attribute_value in attributes.items():
        if isinstance(attribute_value, str):
            if attribute_value.lower().endswith(".csv"):
                attributes[attribute_name] = Path(template_path, attribute_value).__str__()
        elif isinstance(attribute_value, dict):
            replace_timeseries_path_in(attribute_value, template_path)
        elif isinstance(attribute_value, list):
            for item in attribute_value:
                replace_timeseries_path_in(item, template_path)
        else:
            log().debug(DEBUG_NO_PATH_TO_BE_REPLACED_IN.format(attribute_name, attribute_value))


def validate_input_range(input_range: tuple[numeric, numeric], allow_negative: bool) -> None:
    """
    Raises Exception if input range is no int or float or list of [minimum, maximum], and
    values >= 0 (if `allow_negative` = False)
    """
    if (
            isinstance(input_range, tuple)
            and len(input_range) == 2
            and all(isinstance(i, (int, float)) for i in input_range)
    ):
        if not allow_negative:
            if any(i < 0 for i in input_range):
                log_and_raise_critical(ERR_INVALID_RANGE_INPUT.format(input_range, allow_negative))
        if input_range[0] > input_range[1]:
            log_and_raise_critical(ERR_INVALID_RANGE_ORDER.format(input_range, allow_negative))
    else:
        log_and_raise_critical(ERR_INVALID_RANGE_INPUT.format(input_range, allow_negative))


def digest_int_range(input_value: str) -> tuple[int, int]:
    """Returns Tuple of minimum and maximum integer value digested from `input_value` in expected form"""
    numbers = extract_numbers_from_string(input_value, RANGE_INT_IDENTIFIER)
    try:
        min_value, max_value = map(int, numbers.split(SEPARATOR))
        return min_value, max_value
    except ValueError:
        log_and_raise_critical(ERR_COULD_NOT_MAP_RANGE_VALUES.format(numbers))


def digest_float_range(input_value: str) -> tuple[float, float]:
    """Returns Tuple of minimum and maximum float value digested from `input_value` in expected form"""
    numbers = extract_numbers_from_string(input_value, RANGE_FLOAT_IDENTIFIER)
    try:
        min_value, max_value = map(float, numbers.split(SEPARATOR))
        return min_value, max_value
    except ValueError:
        log_and_raise_critical(ERR_COULD_NOT_MAP_RANGE_VALUES.format(numbers))


def digest_choose(input_value: str) -> list[Union[int, float, str]]:
    """Returns List of options digested from given `input_value` string"""
    given_options = (
        input_value.lower()
        .replace(CHOOSE_IDENTIFIER, "")
        .replace("(", "")
        .replace(")", "")
        .replace(" ", "")
        .replace('"', "")
        .replace("'", "")
    )
    given_options = given_options.split(SEPARATOR)
    given_options = cast_numeric_strings(given_options)
    return given_options


def digest_pickfile(input_value: str, scenario_path: Path) -> list[str]:
    """Returns List of all files in given directory `input_value`"""
    relative_path = (
        input_value.lower()
        .replace(PICKFILE_IDENTIFIER, "")
        .replace("(", "")
        .replace(")", "")
        .replace(" ", "")
        .replace('"', "")
        .replace("'", "")
    )
    path_to_dir = Path(scenario_path, relative_path)
    files_in_dir = get_relative_paths_in_dir(path_to_dir, relative_path)
    return files_in_dir


def resolve_identifiers(input_value: Any, options: dict) -> Any:
    """
    Iterates over (potentially nested) `input_value` and returns values from fields
    considering options in `get_value_from_field`
    """
    for key, value in input_value.items():
        if isinstance(value, dict):
            resolve_identifiers(value, options)
        elif isinstance(value, list):
            for index, item in enumerate(value):
                if isinstance(item, dict):
                    resolve_identifiers(item, options)
                else:
                    input_value[key][index] = get_value_from_field(item, options)
        else:
            input_value[key] = get_value_from_field(value, options)


def resolve_ids(scenario: dict) -> None:
    """Resolves in-place all placeholder ID references in Agents & Contracts to unique Ids"""
    active_ids = get_all_ids_from(scenario)
    replacement_map: dict[str, int] = {}
    for agent in scenario["Agents"]:
        agent_id = agent["Id"]
        if REPLACEMENT_IDENTIFIER in str(agent_id):
            if agent_id in replacement_map.keys():
                agent["Id"] = replacement_map[agent_id]
            else:
                unique_id = create_new_unique_id(active_ids)
                agent["Id"] = replacement_map[agent_id] = unique_id
    for contract in scenario["Contracts"]:
        for key, value in contract.items():
            if REPLACEMENT_IDENTIFIER in str(value):
                try:
                    contract[key] = replacement_map[value]
                except KeyError:
                    log_and_raise_critical(ERR_FAILED_RESOLVE_ID.format(value, contract))



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2023 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from scengen.generation.generator import Generator
from scengen.logs import scengen_logger, log
from scengen.cli import arg_handling_run, GeneralOptions, CreateOptions, Command
from scengen.estimator import estimate_scenario
from scengen.files import delete_all_files, increase_count_in_trace_file
from scengen.runner import execute_scenario
from scengen.evaluator import evaluate_scenario


def scengen_cli(args: Optional[list[str]] = None) -> None:
    """Calls sub-commands with appropriate arguments as returned by the command line parser"""
    command, options = arg_handling_run(args)
    scengen_logger(options[GeneralOptions.LOG], options[GeneralOptions.LOGFILE])

    if command is Command.CREATE:
        create(options)


def create(options: dict) -> None:
    """Handles scenario generation based on given `options`"""
    log().info("Starting to create scenarios")
    requested_scenario_count = options[CreateOptions.NUMBER]
    useful_scenario_count = 0
    while useful_scenario_count < requested_scenario_count:
        generator = Generator(options)
        generator.generate_scenarios()

        positive_estimation = True if options[CreateOptions.SKIP_ESTIMATION] else estimate_scenario(options)
        if not positive_estimation:
            log().warning(f"Scenario did not pass estimation. Creating another scenario.")
            delete_all_files(options)
            continue

        execute_scenario(options)

        positive_evaluation = True if options[CreateOptions.SKIP_EVALUATION] else evaluate_scenario(options)
        if positive_evaluation:
            useful_scenario_count += 1
            increase_count_in_trace_file(options)
            log().info(f"Created {useful_scenario_count}/{requested_scenario_count} scenarios.")
        else:
            log().warning(f"Scenario did not pass evaluation. Restarting.")
            delete_all_files(options)
    log().info(f"Created scenario {useful_scenario_count} of {requested_scenario_count}.")


if __name__ == "__main__":
    scengen_cli()



# SPDX-FileCopyrightText: 2024 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import argparse
from enum import Enum, auto
from pathlib import Path
from typing import Any, Optional

from scengen.logs import LogLevel

SCENGEN_PARSER = (
    "Command-line interface to scengen - the scenario generator for the open electricity market model AMIRIS"
)
SCENGEN_LOG_FILE_HELP = "Provide logging file (default: None)"
SCENGEN_LOG_LEVEL_HELP = f"Choose logging level (default: {LogLevel.ERROR.name})"
SCENGEN_COMMAND_HELP = "Choose one of the following commands:"

CREATE_HELP = "Creates scenarios for AMIRIS"
CREATE_N_DEFAULT = 10
CREATE_N_HELP = f"Specify number of scenarios to be generated (default: {CREATE_N_DEFAULT})"
CREATE_CONFIG_HELP = "Path to configuration YAML file defining specifications for creation of scenarios"
CREATE_JAR_HELP = "Path to 'amiris-core_<version>-jar-with-dependencies.jar'"
CREATE_DIR_HELP = "Directory to parse scenarios from and write results to"
CREATE_SKIP_ESTIMATION_HELP = (
    "Speed-focused approach by omitting the AMIRIS scenario estimation at the expense "
    "of bypassing plausibility check (default: False)"
)
CREATE_SKIP_EVALUATION_HELP = (
    "Speed-focused approach by omitting the AMIRIS result evaluation at the expense "
    "of bypassing plausibility check (default: False)"
)
CREATE_OUTPUT_OPTION_HELP = (
    "optional pass through of FAME-Io's output conversion options, see "
    "https://gitlab.com/fame-framework/fame-io/-/blob/main/README.md#read-fame-results"
)
CREATE_NO_CHECK_HELP = "Skip checks for Java installation and correct version"


class GeneralOptions(Enum):
    """Specifies general options for scengen"""
    LOG = auto()
    LOGFILE = auto()


class Command(Enum):
    """Specifies command to execute"""
    CREATE = auto()


class CreateOptions(Enum):
    """Options for command `create`"""
    NUMBER = auto()
    CONFIG = auto()
    JAR = auto()
    DIRECTORY = auto()
    SKIP_ESTIMATION = auto()
    SKIP_EVALUATION = auto()
    OUTPUT_OPTIONS = auto()
    NO_CHECKS = auto()


Options = {
    Command.CREATE: CreateOptions,
}


def arg_handling_run(input_args: Optional[list[str]] = None) -> tuple[Command, dict[Enum, Any]]:
    """Handles command line arguments for `scengen` and returns `command` and its options `args`"""
    parent_parser = argparse.ArgumentParser(prog="amiris-scengen", description=SCENGEN_PARSER)
    parent_parser.add_argument("-lf", "--logfile", type=Path, required=False, help=SCENGEN_LOG_FILE_HELP)
    parent_parser.add_argument(
        "-l",
        "--log",
        default=LogLevel.ERROR.name,
        choices=[level.name.lower() for level in LogLevel],
        help=SCENGEN_LOG_LEVEL_HELP,
    )
    subparsers = parent_parser.add_subparsers(dest="command", required=True, help=SCENGEN_COMMAND_HELP)

    create_parser = subparsers.add_parser("create", help=CREATE_HELP)
    create_parser.add_argument("--number", "-n", type=int, default=CREATE_N_DEFAULT, help=CREATE_N_HELP)
    create_parser.add_argument("--config", "-c", type=Path, required=True, help=CREATE_CONFIG_HELP)
    create_parser.add_argument("--jar", "-j", type=Path, required=True, help=CREATE_JAR_HELP)
    create_parser.add_argument("--directory", "-d", type=Path, default=Path("./"), help=CREATE_DIR_HELP)
    create_parser.add_argument(
        "--skip_estimation", "-ses", default=False, action="store_true", help=CREATE_SKIP_ESTIMATION_HELP
    )
    create_parser.add_argument(
        "--skip_evaluation", "-sev", default=False, action="store_true", help=CREATE_SKIP_EVALUATION_HELP
    )
    create_parser.add_argument("--output-options", "-oo", type=str, default="", help=CREATE_OUTPUT_OPTION_HELP)
    create_parser.add_argument("--no-checks", "-nc", action="store_true", default=False, help=CREATE_NO_CHECK_HELP)

    args = vars(parent_parser.parse_args(input_args))
    command = Command[args.pop("command").upper()]

    args = resolve_relative_paths(args)

    return command, enumify(command, args)


def resolve_relative_paths(args: dict) -> dict:
    """Returns given `args` with relative paths resolved as absolute paths"""
    for option in args:
        if isinstance(args[option], Path):
            args[option] = args[option].resolve()
    return args


def enumify(command: Command, args: dict) -> dict[Enum, Any]:
    """Matches `args` for given `command` to their respective Enum"""
    result = {}
    for option in GeneralOptions:
        result[option] = args.pop(option.name.lower())

    for option in Options[command]:
        result[option] = args.pop(option.name.lower())
    return result



# SPDX-FileCopyrightText: 2023 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import pandas as pd
from amirispy.source.cli import RunOptions

from scengen.logs import log

NAME_ENERGY_EXCHANGE = "DayAheadMarketSingleZone"
NAME_ELECTRICITY_PRICE_COLUMN = "ElectricityPriceInEURperMWH"  # noqa
SCARCITY_PRICE = 3000
THRESHOLD_SHARE_SCARCITY_HOURS = 0.10


def evaluate_scenario(options: dict) -> bool:
    """Returns True if results pass all individual checks"""
    log().debug("Calling evaluator")
    checks = [scarcity_occurrence(options)]
    return all(checks)


def scarcity_occurrence(options: dict) -> bool:
    """Returns True if occurrences of SCARCITY_PRICE is within THRESHOLD_SHARE_SCARCITY_HOURS"""
    path_to_exchange = Path(options[RunOptions.OUTPUT], f"{NAME_ENERGY_EXCHANGE}.csv")
    energy_exchange = pd.read_csv(path_to_exchange, sep=";")
    scarcity_hours = (energy_exchange[NAME_ELECTRICITY_PRICE_COLUMN] >= SCARCITY_PRICE).sum()
    n_of_tolerated_hours = round(len(energy_exchange) * THRESHOLD_SHARE_SCARCITY_HOURS)
    if scarcity_hours > n_of_tolerated_hours:
        decision = False
        log().warning(f"Number of scarcity hours ({scarcity_hours}) exceeds tolerance of {n_of_tolerated_hours} hours.")
    else:
        decision = True
    return decision



import os
import re
from pathlib import Path
from typing import Union


def get_matching_ids_from(scenario: dict, ids_to_look_for: list[Union[str, int]]) -> list:
    """Returns resolved ids of matching `ids_to_look_for` from created agents in `scenario` and static integer ids"""
    resolved_ids = []
    for id_pattern in ids_to_look_for:
        if isinstance(id_pattern, int):
            resolved_ids.append(id_pattern)
        else:
            pattern = re.compile(rf'//{id_pattern}\d*')
            resolved_ids.extend([agent["Id"] for agent in scenario["Agents"] if pattern.match(str(agent["Id"]))])
    return resolved_ids


def get_all_ids_from(scenario: dict) -> list[int]:
    """Returns list of unique Agent Ids in given `scenario`"""
    unique_ids = [agent["Id"] for agent in scenario["Agents"] if isinstance(agent["Id"], int)]
    return list(set(unique_ids))


def create_new_unique_id(unique_ids: list[int]) -> int:
    """Returns new unique integer by adding 1 to the highest Id in `unique_ids`. Adds new Id in-place to `unique_ids`"""
    new_id = max(unique_ids) + 1 if unique_ids else 1
    unique_ids.append(new_id)
    return new_id


def cast_numeric_strings(values: list[str]) -> list[Union[int, float, str]]:
    """Returns given List of `values` but numerics are cast in their correct type"""
    casted_values = []
    for value in values:
        try:
            casted_value = int(value)
        except ValueError:
            try:
                casted_value = float(value)
            except ValueError:
                casted_value = value
        casted_values.append(casted_value)
    return casted_values


def get_relative_paths_in_dir(path_to_dir: Path, relative_path: str) -> list[str]:
    """Returns files in `path_to_dir` with suffix of `relative_path`"""
    files_in_dir = [
        str(Path(relative_path, f)) for f in os.listdir(path_to_dir) if os.path.isfile(os.path.join(path_to_dir, f))
    ]
    return files_in_dir


def extract_numbers_from_string(input_value: str, identifier: str) -> str:
    """Returns string in form 'value, value' from given `input_value` by removing `identifier` and '(' and ')'"""
    extracted_numbers = input_value.lower().replace(identifier, "").replace("(", "").replace(")", "")
    return extracted_numbers



from fameio.input.scenario import Contract
from fameio.tools import ensure_is_list

from scengen.generation.digest import ERR_STATIC_CONTRACT, ERR_MISSING_MATCH, REPLACEMENT_IDENTIFIER
from scengen.logs import log_and_raise_critical


def raise_if_static_contract(contract: Contract) -> None:
    """Raises Error if `contract` is static, e.g. has neither dynamic `sender_id` nor `receiver_id`"""
    sender = str(contract.sender_id)
    receiver = str(contract.receiver_id)
    identifier_missing = REPLACEMENT_IDENTIFIER not in sender and REPLACEMENT_IDENTIFIER not in receiver
    if identifier_missing:
        log_and_raise_critical(ERR_STATIC_CONTRACT.format(contract))


def raise_if_dynamic_match_missing(create_config: list[dict]) -> None:
    """Raises Error if any external id(s) cannot be matched to dynamically created agents - skips integer ids"""
    agent_aliases = [config["this_agent"] for config in create_config]
    for config in create_config:
        missing_ids = []
        for external_id_key, external_id_values in config.get("external_ids", {}).items():
            for entry in ensure_is_list(external_id_values):
                if entry not in agent_aliases and isinstance(entry, str):
                    missing_ids.append(entry)
        if missing_ids:
            log_and_raise_critical(ERR_MISSING_MATCH.format(missing_ids, agent_aliases))



# SPDX-FileCopyrightText: 2023 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import os
import shutil
from pathlib import Path
import time

import yaml
from fameio.input.loader import load_yaml

from scengen.cli import CreateOptions
from scengen.logs import log_and_raise_critical, log_error_and_raise, log

_ERR_NOT_A_FOLDER = "Given Path '{}' is not a directory."
_INFO_NO_TRAC_FILE_FOUND = ("Could not find `trace_file` in path '{}' as specified in GeneratorConfig. "
                            "Created new one instead.")
_WARN_NO_TRACE_FILE_DEFINED = ("No mandatory `trace_file` found in given GeneratorConfig. Created new one at '{}' and "
                               "added to GeneratorConfig instead.")


def delete_all_files(options: dict):
    """Removes all created files of scenario stored in `CreateOptions.DIRECTORY`"""
    dir_to_remove = Path(options[CreateOptions.DIRECTORY], options["scenario_name"])
    shutil.rmtree(dir_to_remove, ignore_errors=True)
    os.remove(options["scenario_path"])
    log().debug(f"Removed all files in '{dir_to_remove}'")


def increase_count_in_trace_file(options: dict) -> None:
    """Increases count in trace file defined in `CreateOptions.CONFIG` by 1"""
    config = load_yaml(options[CreateOptions.CONFIG])
    base_path = Path(options[CreateOptions.CONFIG]).parent
    full_path = Path(base_path, config["defaults"]["trace_file"])
    trace_file = load_yaml(full_path)
    trace_file["total_count"] += 1
    write_dict_to_disk(trace_file, full_path)
    log().debug(f"Increased trace file count to '{trace_file['total_count']}'")


def write_dict_to_disk(trace_file: dict, file_name: Path) -> None:
    """Writes `trace_file` as `file_name` to disk"""
    with open(file_name, "w") as file:
        yaml.dump(trace_file, file, default_flow_style=False)


def save_seed_to_trace_file(options: dict, seed: int) -> None:
    """Saves seed to trace file"""
    config = load_yaml(options[CreateOptions.CONFIG])
    base_path = Path(options[CreateOptions.CONFIG]).parent
    full_path = Path(base_path, config["defaults"]["trace_file"])
    trace_file = load_yaml(full_path)
    trace_file["seed"] = seed
    write_dict_to_disk(trace_file, full_path)
    log().debug(f"Stored seed '{seed}' to `trace_file`")


def write_yaml(output_file: dict, output_file_path: Path) -> None:
    """Writes given `output_file` to `output_file_path` (ending in .yaml)"""
    ensure_folder_exists(output_file_path.parent)
    check_if_valid_yaml_path(output_file_path)
    with open(output_file_path, "w") as stream:
        try:
            yaml.safe_dump(output_file, stream)
        except yaml.YAMLError:
            log_and_raise_critical(f"Failed writing YAML to '{output_file_path}'")


def check_if_valid_yaml_path(file_path: Path):
    """Raises Exception if provided `file_path` is not a valid YAML path"""
    if Path(file_path).suffix.lower() not in [".yaml", ".yml"]:
        log_and_raise_critical(f"Provide a valid Path to a YAML file. Received '{file_path}'")


def ensure_folder_exists(path: Path) -> None:
    """
    Returns Path to a directory and creates the folder if required.
    If given Path is an existing folder: does nothing, else creates new folder (including parent folders)

    Args:
        path: to check and create if not existing

    Returns:
        None

    Raises:
        ValueError: if path is an existing file
    """
    if path.is_file():
        log_error_and_raise(ValueError(_ERR_NOT_A_FOLDER.format(path)))
    if not path.is_dir():
        path.mkdir(parents=True)


def get_trace_file(config: dict, options: dict) -> dict:
    """Returns loaded `defaults["trace_file"]` or creates new trace file if not present"""
    defaults = config["defaults"]
    if defaults.get("trace_file"):
        trace_file_path = defaults["trace_file"]
        try:
            trace_file = load_yaml(Path(options[CreateOptions.CONFIG].parent, trace_file_path))
        except FileNotFoundError:
            trace_file = setup_new_trace_file(config, defaults, options, trace_file_path)
            log().info(_INFO_NO_TRAC_FILE_FOUND.format(trace_file_path))
    else:
        trace_file_name = f"trace_file_{time.strftime('%Y-%m-%d_%H%M%S')}.yaml"
        trace_file = setup_new_trace_file(config, defaults, options, trace_file_name)
        log().warning(_WARN_NO_TRACE_FILE_DEFINED.format(trace_file_name))
    return trace_file


def setup_new_trace_file(config: dict, defaults: dict, options: dict, file_name: str) -> dict:
    """Returns new trace_file which is added to config and saved to disk with `file_name`"""
    trace_file = {
        "total_count": 0,
        "seed": 0,
    }

    defaults["trace_file"] = file_name
    write_dict_to_disk(config, options[CreateOptions.CONFIG])

    base_path = Path(options[CreateOptions.CONFIG]).parent
    full_path = Path(base_path, file_name)
    ensure_folder_exists(full_path.parent)
    write_dict_to_disk(trace_file, full_path)
    return trace_file



# SPDX-FileCopyrightText: 2023 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
from fameio.input.loader import load_yaml

from scengen.logs import log_and_raise_critical, log


class Amiris:
    capacitiy_name = "InstalledPowerInMW"
    energy_carrier_name = "EnergyCarrier"
    fuel_type_name = "FuelType"
    prototype_name = "Prototype"
    technology_name_for_storage = "Storage"
    identifier_for_storage = "Device"
    capacitiy_name_for_storage = "InstalledPowerInMW"


def get_installed_capacity(scenario: dict) -> dict:
    """Returns `dict` of technologies each with list of {agent_id: installed capacity} from given `scenario`"""
    installed_power_by_type = dict()
    for agent in scenario["Agents"]:
        if "Attributes" in agent:
            if Amiris.capacitiy_name in agent["Attributes"]:
                capacity, technology = extract_conventional(agent)
                installed_power_by_type.setdefault(technology, []).append({agent["Id"]: capacity})
            if Amiris.identifier_for_storage in agent["Attributes"]:
                capacity, technology = extract_storage(agent)
                installed_power_by_type.setdefault(technology, []).append({agent["Id"]: capacity})
            if "Plants" in agent["Attributes"]:
                for plant in agent["Attributes"]["Plants"]:
                    capacity = plant["NetCapacityInMW"]
                    technology = agent["Attributes"]["Prototype"]["FuelType"]
                    if "Id" in plant:
                        plant_id = int(plant["Id"])
                        installed_power_by_type.setdefault(technology, []).append({plant_id: capacity})
                    elif capacity > 0:
                        log().warning("Missing `Id` for powerplant with power of {}".format(capacity))
    return installed_power_by_type


def extract_storage(agent: dict) -> tuple[float, str]:
    """Return capacity and technology of storage `agent`"""
    capacity = agent["Attributes"][Amiris.identifier_for_storage][Amiris.capacitiy_name_for_storage]
    technology = Amiris.technology_name_for_storage
    return capacity, technology


def extract_conventional(agent: dict) -> tuple[float, str]:
    """Return capacity and technology of conventional `agent`"""
    capacity = agent["Attributes"][Amiris.capacitiy_name]
    if Amiris.energy_carrier_name in agent["Attributes"]:
        technology = agent["Attributes"][Amiris.energy_carrier_name]
    elif Amiris.fuel_type_name in agent["Attributes"][Amiris.prototype_name]:
        technology = agent["Attributes"][Amiris.prototype_name][Amiris.fuel_type_name]
    else:
        log_and_raise_critical("Found no eligible technology for agent '{}' with id '{}'".format(agent, agent["Id"]))
    return capacity, technology


def accumulate_capacities(capacities: dict) -> float:
    """Returns accumulates `installed_capacities` in MW as float"""
    installed_capacities = 0
    for _, technology in capacities.items():
        for item in technology:
            for _, MW in item.items():
                installed_capacities += MW
    return installed_capacities


def generation_capacity_available(scenario: dict) -> bool:
    """Returns True if there are any installed capacities in given `scenario`"""
    capacities = get_installed_capacity(scenario)
    accumulated_capacities = accumulate_capacities(capacities)
    if not accumulated_capacities > 0:
        log().warning(f"Accumulated installed capacities seems very low at '{accumulated_capacities}' MW")
        decision = False
    else:
        decision = True
    return decision


def estimate_scenario(options: dict) -> bool:
    """Returns True if scenario passes all individual checks"""
    log().debug("Calling estimator")
    scenario = load_yaml(options["scenario_path"])
    checks = [generation_capacity_available(scenario)]
    return all(checks)



import copy
import random
import time
from pathlib import Path

from fameio.input.loader import load_yaml
from fameio.input.scenario import Contract
from fameio.tools import ensure_is_list

from scengen.cli import CreateOptions
from scengen.files import get_trace_file, save_seed_to_trace_file, write_yaml
from scengen.generation.digest import get_number_of_agents_to_create, get_agent_id, \
    update_series_paths, resolve_identifiers, resolve_ids, KEY_THIS_AGENT, REPLACEMENT_IDENTIFIER
from scengen.generation.check import raise_if_static_contract, raise_if_dynamic_match_missing
from scengen.generation.misc import get_matching_ids_from
from scengen.logs import log


DEBUG_NO_CREATE = "No agents to `create` found in Config '{}'"


class Generator:
    """Main scenario generator class"""
    def __init__(self, options: dict):
        self.options = options
        self.config = load_yaml(self.options[CreateOptions.CONFIG])
        self.trace_file = get_trace_file(self.config, options)
        self.scenario = {}

    def generate_scenarios(self) -> None:
        """Generates a new scenario based on `options` and `scenario_name` stored in `CreateOptions.DIRECTORY`"""
        log().debug("Generating scenario")
        self._init_random_seed()

        defaults = self.config["defaults"]
        count = self.trace_file["total_count"]
        self._set_scenario_name(defaults["base_name"] + f"_{count}")
        self._set_scenario(load_yaml(Path(self.options[CreateOptions.CONFIG].parent, self.config["base_template"])))

        if "create" in self.config:
            raise_if_dynamic_match_missing(self.config["create"])
            self.add_agents()
            self.add_contracts()
        else:
            log().debug(DEBUG_NO_CREATE)

        resolve_identifiers(self.scenario, self.options)
        resolve_ids(self.scenario)
        update_series_paths(self.scenario, self.options, Path(self.config["base_template"]))

        self._set_scenario_path(Path(self.options[CreateOptions.DIRECTORY], self.options["scenario_name"] + ".yaml"))
        write_yaml(self.scenario, self.options["scenario_path"])

    def _init_random_seed(self) -> None:
        """Initializes random seed if not yet saved to `options['random_seed']`"""
        if not self.options.get("random_seed"):
            random_seed = self._get_random_seed()
            random.seed(random_seed + self.trace_file["total_count"])
            self.options["random_seed"] = self.trace_file["random_seed"] = random_seed
            save_seed_to_trace_file(self.options, random_seed)

    def _get_random_seed(self) -> int:
        """Returns random seed as integer, defined optionally in `defaults['seed']` or from system time in ns instead"""
        defaults = self.config["defaults"]
        if defaults.get("seed"):
            random_seed = defaults["seed"]
        else:
            random_seed = time.time_ns()
        return random_seed

    def _set_scenario_name(self, scenario_name: str) -> None:
        """Sets `scenario_name` in `options`"""
        self.options["scenario_name"] = scenario_name

    def _set_scenario(self, scenario: dict) -> None:
        """Sets `scenario`"""
        self.scenario = scenario

    def _set_scenario_path(self, path: Path) -> None:
        """Sets `scenario_name` in `options`"""
        self.options["scenario_path"] = path

    def add_agents(self) -> None:
        """Adds agents to create to `scenario`"""
        for agent in self.config["create"]:
            type_template = load_yaml(Path(self.options[CreateOptions.CONFIG].parent, agent["type_template"]))
            agent_type_template = type_template["Agent"]
            n_to_create = get_number_of_agents_to_create(agent["count"], self.options)
            agent_name = agent["this_agent"]

            for n in range(n_to_create):
                agent_to_append = copy.deepcopy(agent_type_template)
                agent_to_append["Id"] = get_agent_id(agent_name, n, n_to_create)
                self.scenario["Agents"].append(agent_to_append)

    def add_contracts(self) -> None:
        """
        Iterates over agents and adds contracts dynamically to `scenario`.
        Note: Ensure that all dynamic agents are already added to the scenario
        """
        for agent in self.config["create"]:
            type_template = load_yaml(Path(self.options[CreateOptions.CONFIG].parent, agent["type_template"]))
            if not type_template.get("Contracts"):
                continue
            id_map = {KEY_THIS_AGENT: get_matching_ids_from(self.scenario, ensure_is_list(agent["this_agent"]))}
            for id_key, id_values in agent.get("external_ids", {}).items():
                matched_ids = get_matching_ids_from(self.scenario, ensure_is_list(id_values))
                id_map[REPLACEMENT_IDENTIFIER + id_key] = matched_ids
            for contract in type_template.get("Contracts"):
                contract = Contract.from_dict(contract)
                raise_if_static_contract(contract)
                contracts_to_append = self._create_contracts(contract, id_map)
                self.scenario["Contracts"].extend(contracts_to_append)

    @staticmethod
    def _create_contracts(contract: Contract, id_map: dict) -> list[dict]:
        """Returns list of dynamically created `contracts` based on template `contract` and `id_map`"""
        created_contracts = []
        sender_default = str(contract.sender_id)
        receiver_default = str(contract.receiver_id)
        for sender_override in id_map.get(sender_default, [sender_default]):
            for receiver_override in id_map.get(receiver_default, [receiver_default]):
                contract_to_append = copy.deepcopy(contract.to_dict())
                if REPLACEMENT_IDENTIFIER in sender_default:
                    # noinspection PyProtectedMember
                    contract_to_append[Contract.KEY_SENDER] = sender_override
                if REPLACEMENT_IDENTIFIER in receiver_default:
                    # noinspection PyProtectedMember
                    contract_to_append[Contract.KEY_RECEIVER] = receiver_override
                created_contracts.append(contract_to_append)
        return created_contracts



# SPDX-FileCopyrightText: 2024 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import os
from pathlib import Path

from amirispy.scripts.subcommands import run as amiris
from amirispy.source.cli import GeneralOptions as AMIRISGeneralOptions

from scengen.cli import CreateOptions, GeneralOptions
from scengen.logs import log

NAME_SCENARIO_YAML = "scenario.yaml"


def execute_scenario(options: dict) -> None:
    """Calls AMIRIS after mapping `options` using amirispy functionality"""
    log().debug("Executing scenario")
    options = map_options(options)
    amiris.run_amiris(options)
    delete_pb_files()


def map_options(options: dict) -> dict:
    """Maps values from scengen `options` to option keys of amirispy"""
    options[amiris.RunOptions.JAR] = options[CreateOptions.JAR]
    options[amiris.RunOptions.OUTPUT] = Path(options[CreateOptions.DIRECTORY], options['scenario_path'].stem)
    options[AMIRISGeneralOptions.LOG] = options[GeneralOptions.LOG]
    options[AMIRISGeneralOptions.LOGFILE] = options[GeneralOptions.LOGFILE]
    options[amiris.RunOptions.SCENARIO] = options["scenario_path"]
    options[amiris.RunOptions.OUTPUT_OPTIONS] = options[CreateOptions.OUTPUT_OPTIONS]
    options[amiris.RunOptions.NO_CHECKS] = options[CreateOptions.NO_CHECKS]
    return options


def delete_pb_files() -> None:
    """Deletes input and output Protobuf files created by fameio in current working dir"""
    for file in ["input.pb", "output.pb"]:
        try:
            os.remove(Path(os.getcwd(), file))
        except FileNotFoundError:
            log().debug(f"Could not delete file `{file}` in current working dir.")
